import os

from tests.input_files.mypackage.cmodule import ExampleClass
from tests.test_pipeline_manager.test_create import PipelineManagerLoadTestBase

from pyfileconf import Selector
from tests.input_files.amodule import a_function
from tests.test_pipeline_manager.base import PipelineManagerTestBase, CLASS_CONFIG_DICT_LIST


class PipelineManagerCreateEntryTestBase(PipelineManagerLoadTestBase):
    pass


class TestPipelineManagerCreateEntry(PipelineManagerCreateEntryTestBase):

    def test_create_entry_for_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing', a_function)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.a_function
        module_folder = os.path.join(self.defaults_path, 'thing')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)

    def test_create_entry_for_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing', ExampleClass)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.ExampleClass
        module_folder = os.path.join(self.defaults_path, 'thing')
        class_path = os.path.join(module_folder, 'ExampleClass.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_config_file_contents(contents)

    def test_create_entry_for_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.thing.data')
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.thing.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'thing')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)

    def test_create_deeply_nested_entry_for_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing.stuff.whoa', a_function)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.stuff.whoa.a_function
        module_folder = os.path.join(self.defaults_path, 'thing')
        module_folder = os.path.join(module_folder, 'stuff')
        module_folder = os.path.join(module_folder, 'whoa')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)

    def test_create_deeply_nested_entry_for_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.thing.stuff.whoa.data')
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.thing.stuff.whoa.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'thing')
        module_folder = os.path.join(module_folder, 'stuff')
        module_folder = os.path.join(module_folder, 'whoa')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)

    def test_create_entry_for_specific_class_dict_in_same_section_as_existing_entry(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.stuff.data2')
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data2
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'stuff')
        class_path = os.path.join(module_folder, 'data2.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(
                contents,
                name_value="name: Optional[str] = 'data2'"
            )



from typing import List, Union
import os
import warnings

from mixins.repr import ReprMixin
from pyfileconf.basemodels.container import Container
from pyfileconf.config.models.config import ActiveFunctionConfig
from pyfileconf.exceptions.config import ConfigManagerNotLoadedException
from pyfileconf.sectionpath.sectionpath import _strip_py


class ConfigSection(Container, ReprMixin):
    repr_cols = ['name', 'config', 'items']

    def __init__(self, configs: List[Union[ActiveFunctionConfig, 'ConfigSection']],
                 section_config: ActiveFunctionConfig=None, name: str=None):
        self.config = section_config
        self.items = configs
        self.name = name

    def __getattr__(self, item):

        # Handle if trying to get this section's config
        if self.config is not None and item == self.config.name:
            return self.config

        return self.config_map[item]

    def __dir__(self):
        return self.config_map.keys()

    @property  # type: ignore
    def items(self):
        return self._items

    @items.setter
    def items(self, items):
        self._items = items
        self._set_config_map() # recreate config map on setting items

    @property
    def config_map(self):
        return self._config_map

    def _set_config_map(self):
        config_map = {}
        for config in self:
            if config.name is None:
                warnings.warn(f"Couldn't determine name of config {config}. Can't add to mapping.")
                continue
            config_map[config.name] = config
        self._config_map = config_map

    def update(self, d: dict=None, **kwargs):
        if self.config is None:
            raise ConfigManagerNotLoadedException('no config in ConfigSection')
        if d is not None:
            self.config.update(d)
        self.config.update(kwargs)

    @classmethod
    def from_files(cls, basepath: str):

        # Get section name by name of folder
        name = os.path.basename(os.path.abspath(basepath))

        # Config folder doesn't exist, return empty config
        if not os.path.exists(basepath):
            return cls([], name=name)

        config_file_list = [file for file in next(os.walk(basepath))[2] if file.endswith('.py')]

        # Ignore folders starting with . or _
        config_section_list = [
            folder for folder in next(os.walk(basepath))[1] if not any(
                [folder.startswith(item) for item in ('_','.')]
            )
        ]

        # Special handling for section config
        try:
            config_file_list.remove('section.py')
            section_config = ActiveFunctionConfig.from_file(os.path.join(basepath, 'section.py'), name=name)
        except ValueError:
            # Didn't find section config
            section_config = None

        configs = [
            ActiveFunctionConfig.from_file(
                os.path.join(basepath, file),
                name=_strip_py(file)
            ) for file in config_file_list
        ]
        # Recursively calling section creation to create individual config files
        config_sections = [ConfigSection.from_files(os.path.join(basepath, folder)) for folder in config_section_list]

        configs += config_sections

        return cls(configs, section_config=section_config, name=name)

